-------------------------------------------------------------------------------
--
-- Copyright 2022
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------
-- Author: Job van Wee
-- Purpose: Self checking and self-stopping tb for ddrctrl_input_address_counter.vhd
-- Usage:
-- > run -a

LIBRARY IEEE, common_lib, technology_lib, tech_ddr_lib, dp_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE technology_lib.technology_pkg.ALL;
USE tech_ddr_lib.tech_ddr_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_pkg.ALL;

ENTITY tb_ddrctrl_input_address_counter IS
  GENERIC (
    
    g_tech_ddr                : t_c_tech_ddr    := c_tech_ddr4_8g_1600m;                    -- type of memory
    g_sim_model               : BOOLEAN         := TRUE;                                    -- determens if this is a simulation
    g_sim_length              : NATURAL         := 52                                       -- determens the length of the duration of the test

  );
END tb_ddrctrl_input_address_counter;

ARCHITECTURE tb OF tb_ddrctrl_input_address_counter IS

  -- constants for running the testbench
  CONSTANT  c_clk_freq        : NATURAL         := 200;                                     -- clock frequency in MHz
  CONSTANT  c_clk_period      : TIME            := (10**6/c_clk_freq)*1 ps;                 -- clock period, 5 ns

  -- constants for running the test
  CONSTANT  c_data_w          : NATURAL         := func_tech_ddr_ctlr_data_w( g_tech_ddr ); -- in and output data vector with, 576
  CONSTANT  c_adr_w           : NATURAL         := func_tech_ddr_ctlr_address_w( g_tech_ddr ); -- address with in simulation
  CONSTANT  c_adr_size        : NATURAL         := 2**c_adr_w;                              -- address size in simulation

  -- input signals for ddrctrl_address_counter.vhd
  SIGNAL    clk               : STD_LOGIC       := '1';
  SIGNAL    rst               : STD_LOGIC       := '0';
  SIGNAL    q_rst             : STD_LOGIC       := '0';
  SIGNAL    q_q_rst           : STD_LOGIC       := '0';
  SIGNAL    in_sosi           : t_dp_sosi       := c_dp_sosi_init;                          -- signal which is the input for ddrctrl_address_counter.vhd
  SIGNAL    in_of             : NATURAL         := 0;                                       -- signal which contains the amount of overflow

  -- output signal from ddrctrl_address_counter.vhd
  SIGNAL    out_sosi          : t_dp_sosi := c_dp_sosi_init;                     -- signal which is the output from ddrctrl_address_counter.vhd
  SIGNAL    out_adr           : NATURAL         := 0;
  SIGNAL    out_of            : NATURAL         := 0;                                       -- signal which is the output from ddrctrl_address_counter.vhd

  -- testbench signals
  SIGNAL    tb_end            : STD_LOGIC       := '0';                                     -- signal to turn the testbench off
  SIGNAL    in_data           : STD_LOGIC_VECTOR(c_data_w-1 DOWNTO 0) := (OTHERS => '0');   -- signal which contains the data that is set as input
  SIGNAL    q_in_data         : STD_LOGIC_VECTOR(c_data_w-1 DOWNTO 0) := (OTHERS => '0');   -- signal which contains the data that is set as input with a delay of 1 clockcycle
  SIGNAL    q_q_in_data       : STD_LOGIC_VECTOR(c_data_w-1 DOWNTO 0) := (OTHERS => '0');   -- signal which contains the data that is set as input with a delay of 2 clockcycles
  SIGNAL    in_data_enable    : STD_LOGIC       := '0';                                     -- signal to determen if in_data is ready for reading
  SIGNAL    q_in_data_enable  : STD_LOGIC       := '0';                                     -- signal to determen if in_data is ready for reading with a delay of 1 clockcycle
  SIGNAL    q_q_in_data_enable: STD_LOGIC       := '0';                                     -- signal to determen if in_data is ready for reading with a delay of 2 clockcycles
  SIGNAL    q_in_of           : NATURAL         := 0;                                       -- signal which contains the amount of overflow with a delay of 1 clockcycle
  SIGNAL    q_q_in_of         : NATURAL         := 0;                                       -- signal which contains the amount of overflow with a delay of 2 clockcycles
  SIGNAL    lag_due_reset     : NATURAL         := 0;                                       -- signal to hold the address lag after a reset
  SIGNAL    q_lag_due_reset   : NATURAL         := 0;                                       -- signal to hold the address lag after a reset with a delay of 1 clockcycle

BEGIN

  -- wiring the input signals to the inputs of the testbench
  in_sosi.data(c_data_w-1 DOWNTO 0) <= in_data(c_data_w-1 DOWNTO 0);
  in_sosi.valid                     <= in_data_enable;

  -- generating clock
  clk <= NOT clk OR tb_end AFTER c_clk_period/2;

  -- excecuting the test
  p_test : PROCESS
  BEGIN

    -- initialize inputs
    tb_end            <= '0';
    in_data           <= (OTHERS => '0');
    in_data_enable    <= '0';
    WAIT UNTIL rising_edge(clk);
    WAIT FOR c_clk_period*4;
    rst <= '1';
    WAIT FOR c_clk_period*1;
    rst <= '0';

    -- changing inputs to start the address counting
    FOR I IN 0 TO g_sim_length-1 LOOP
      in_data_enable  <= '1';
      in_data         <= NOT in_data;
      in_of           <= in_of + 1;
      WAIT FOR c_clk_period*1;
      in_data_enable  <= '0';
      WAIT FOR c_clk_period*2;
    END LOOP;

    -- stopping the test
    WAIT FOR c_clk_period*4;
    tb_end            <= '1';
    ASSERT FALSE                                                                      REPORT "Test: OK"                                                                                                                           SEVERITY FAILURE;
  END PROCESS;

  -- generating compare data for out_sosi
  p_out_sosi : PROCESS
  BEGIN
    WAIT UNTIL rising_edge(clk);
    if rising_edge(clk) THEN
      q_q_in_data_enable <= q_in_data_enable;
      q_q_in_data        <= q_in_data;
      q_q_in_of          <= q_in_of;
      q_q_rst            <= q_rst;
      q_in_data_enable   <= in_data_enable;
      q_in_data          <= in_data;
      q_in_of            <= in_of;
      q_lag_due_reset    <= lag_due_reset;
      q_rst              <= rst;
    END IF;
    IF q_rst = '1' THEN
      IF lag_due_reset + out_adr >= c_adr_size THEN
        lag_due_reset <= lag_due_reset+out_adr-c_adr_size;
      ELSE
        lag_due_reset <= lag_due_reset+out_adr;
      END IF;
    END IF;
  END PROCESS;
      

  -- verifying if the data is correct and if valid is correct
  p_verify_data_valid : PROCESS
  BEGIN
    WAIT UNTIL rising_edge(clk);
    IF rising_edge(clk) THEN
      ASSERT q_q_in_data(c_data_w-1 DOWNTO 0) = out_sosi.data(c_data_w-1 DOWNTO 0)    REPORT "in_sosi.data does not match out_sosi.data"                                                                                          SEVERITY ERROR;
      ASSERT q_q_in_data_enable               = out_sosi.valid                        REPORT "in_sosi.valid does not match out_sosi.valid"                                                                                        SEVERITY ERROR;
      ASSERT q_q_in_of                        = out_of                                REPORT "in_of does not match out_of"                                                                                                        SEVERITY ERROR;
    END IF;
  END PROCESS;

  -- excecuting the reset test
  p_test_reset : PROCESS
  BEGIN
    rst <= '0';
    WAIT FOR c_clk_period*((c_adr_size*4)+3);
    rst <= '1';
    WAIT FOR c_clk_period*1;
  END PROCESS;

  -- verifying if the address is correct by keeping track of the address
  p_verify_address : PROCESS

  VARIABLE v_adr  : NATURAL range 0 to c_adr_size-1 := c_adr_size-1;

  BEGIN
    WAIT UNTIL rst = '1';
    WAIT UNTIL rst = '0';
    FOR I IN 0 TO g_sim_length-1 LOOP
      IF v_adr >= q_lag_due_reset THEN
        ASSERT v_adr-q_lag_due_reset              = out_adr                      REPORT "Wrong address, 1, v_adr = " & NATURAL'image(v_adr-q_lag_due_reset) & ", address = " & NATURAL'image(out_adr)                       SEVERITY ERROR;
      ELSE
        ASSERT (v_adr-q_lag_due_reset)+c_adr_size = out_adr                      REPORT "Wrong address, 2, v_adr = " & NATURAL'image((v_adr-q_lag_due_reset)+c_adr_size) & ", address = " & NATURAL'image(out_adr)          SEVERITY ERROR;
      END IF;
      WAIT UNTIL out_sosi.valid = '1';     
      IF q_q_rst = '1' THEN
        WAIT UNTIL out_sosi.valid = '1';
      END IF;
      IF v_adr = c_adr_size-1 THEN
        v_adr := 0;
      ELSE
        v_adr := v_adr+1;
      END IF;
    END LOOP;
  END PROCESS;


  -- DUT
  u_ddrctrl_input_address_counter : ENTITY work.ddrctrl_input_address_counter
  GENERIC MAP (
    g_tech_ddr        => g_tech_ddr,
    g_sim_model       => g_sim_model
  )
  PORT MAP (
    clk               => clk,
    rst               => rst,
    in_sosi           => in_sosi,
    in_of             => in_of,

    out_sosi          => out_sosi,
    out_of            => out_of,
    out_adr           => out_adr
  );

END tb;
